# -*- coding: utf-8 -*-
"""
    Jinja24doc is lightweight documentation generator for python modules with
    jinja2 templates. It is part of *Poor Http* group tools (WSGI connector,
    WSGI/HTTP Server and mod_python connector). It could load modules and gets
    documentation for its items. No configuration is needed, only jinja2
    templates. Your or from jinja2doc package.
"""

__author__ = "Ondřej Tůma (McBig) <mcbig@zeropage.cz>"
__date__ = "29 Jul 2019"
__version__ = "1.3.3"



"""Library use context object"""

from jinja2 import Environment, FileSystemLoader

from sys import stderr
from os import path, getcwd

from jinja24doc.apidoc import local_name, property_info
from jinja24doc.wiki import Wiki
from jinja24doc.rst import Rst

TEMPLATES = "../../../../share/jinja24doc/templates/"


def log(message):
    """
    Write message to stderr.

        #!jinja
        {% do log('some debug message') %}
    """
    stderr.write("%s\n" % message)


class Context(Wiki, Rst):

    def __init__(self, paths, encoding='utf-8'):
        super(Context, self).__init__()
        if isinstance(paths, tuple):
            paths = list(paths)
        assert isinstance(paths, list), "Paths is %s" % paths.__class__
        self.paths = paths + [getcwd(),
                              path.abspath(path.join(path.dirname(__file__),
                                                     TEMPLATES))]
        self.encoding = encoding

    def prepare_environment(self):
        """Prepare jinja2 environment.

        This method is called internal by Context.generate method, and append
        Conetxt methods to jinja2 template globals. So Context.load_module,
        Context.keywords, local_name, property_info, wiki, Context.load_wiki,
        Context.load_text, Context.load_source, Context.rst, Context.load_rst
        and log methods and functions are enabled to call in templates.
        """
        env = Environment(loader=FileSystemLoader(self.paths),
                          trim_blocks=True,
                          extensions=['jinja2.ext.do',
                                      'jinja2.ext.loopcontrols'],
                          lstrip_blocks=True)     # add in 2.7

        env.globals['load_module'] = self.load_module
        env.globals['keywords'] = self.keywords
        env.globals['local_name'] = local_name
        env.globals['property_info'] = property_info

        env.globals['wiki'] = self.wiki
        env.globals['load_wiki'] = self.load_wiki
        env.globals['load_text'] = self.load_text
        env.globals['load_source'] = self.load_source

        env.globals['rst'] = self.rst
        env.globals['load_rst'] = self.load_rst
        env.globals['log'] = log
        return env

    def generate(self, template, **kwargs):
        """Generate html output from template."""
        env = self.prepare_environment()
        temp = env.get_template(template)
        return temp.render(**kwargs)




"""
Library for reStrucuredText parsing, and generating simple HTML output.
"""

from docutils.core import publish_parts
from docutils_tinyhtml import Writer

from sys import version_info
if version_info[0] == 2:
    from io import open

import os

from jinja24doc.apidoc import ApiDoc, pep_rfc
from jinja24doc.wiki import re_source, re_python, _python


def _doctest_code(obj):
    tmp = obj.groups()
    if tmp[0] == 'doctest':
        source = re_python.sub(_python, tmp[1])
    else:
        source = tmp[1]
    return '<pre class="%s">%s</pre>' % (tmp[0], source)


class Rst(ApiDoc):
    """Class based on ApiDoc."""

    def rst(self, doc, link='link', top='top', title='__doc__',
            section_level=2, system_message=False):
        """
        Call reStructuredText docutil parser for doc and return it with html
        representation of reStructuredText formating. For more details see
        http://docutils.sourceforge.net/rst.html.
        """
        writer = Writer()
        parts = publish_parts(source=doc,
                              writer=writer,
                              writer_name='html',
                              settings_overrides={
                                  'link': link, 'top': top, 'title': title,
                                  'initial_header_level': section_level,
                                  'no_system_messages': not system_message})

        out = parts['body'] + parts['html_line'] + \
            parts['html_footnotes'] + parts['html_citations']

        out = re_source.sub(_doctest_code, out.strip())

        if out.startswith('<p') and out.endswith('</p>') and \
                out.count('</p>') == 1:     # strip paragraph if is one
            out = out[out.index('>')+1:-4]
        return self.linked_api(pep_rfc(out))

    def load_rst(self, rstfile, link='link', top='top', system_message=False):
        """Load reStructuredText file and create docs list of headers and text.

        Parameters:
            rstfile - string, reStructured source file name (readme.rst)
            link - link label for headers. If is empty, link href will be
                hidden.
            top - top label for headers. If is empty, top href will be hidden.

            #!jinja
            {% set sections = load_rst('readme.rst', '', '') %}
            {% type, filename, _none_, text = sections[-1] %}
        """

        x_rstfile = ''
        for path in self.paths:
            if os.access(path+'/'+rstfile, os.R_OK):
                x_rstfile = path+'/'+rstfile
                break
        if not x_rstfile:
            raise SystemExit('Access denied to text file %s' % rstfile)

        with open(x_rstfile, 'r', encoding=self.encoding) as f:
            doc = f.read()

        writer = Writer()
        parts = publish_parts(source=doc,
                              source_path=x_rstfile,
                              writer=writer,
                              writer_name='html',
                              settings_overrides={
                                  'link': link, 'top': top,
                                  'no_system_messages': not system_message})

        out = parts['body']
        if parts['html_footnotes'] or parts['html_citations']:
            out += parts['html_line'] + \
                parts['html_footnotes'] + parts['html_citations']

        out = re_source.sub(_doctest_code, out)

        if out.startswith('<p') and out.endswith('</p>') and \
                out.count('</p>') == 1:     # strip paragraph if is one
            out = out[out.index('>')+1:-4]

        out = self.linked_api(pep_rfc(out))

        retval = list(('h%d' % lvl, self.uni(name), id, '')
                      for lvl, name, id in parts['sections'])
        # TODO: append '(author, date, verstion)' from rst if exist like in
        # module
        retval.append(('text', parts['title'], None, self.uni(out)))
        return retval
# endclass



""" module api documentation reader """

from inspect import getargspec, getdoc, getmembers, getsource, formatargspec, \
    isfunction, isroutine, ismethod, isclass, ismodule, isbuiltin, \
    ismethoddescriptor, isgetsetdescriptor
from operator import itemgetter

import sys
import re

# from Python 3.7 re.Pattern type instead of re._pattern_type
try:
    from re import _pattern_type as Pattern
except ImportError:
    from re import Pattern


unicode_exist = True
try:
    unicode()
except:
    unicode_exist = False


ordering = {'module':       (0, 0),
            'submodule':    (1, 0),
            'dependence':   (1, 1),
            'class':        (2, 0),
            'property':     (2, 1),
            'descriptor':   (2, 2),
            'method':       (2, 3),
            'staticmethod': (2, 4),
            'function':     (3, 0),
            'variable':     (4, 0)}

re_notlink = re.compile(r"(.*?)((<a .*?</a>)|$)", re.S)

re_pep = re.compile(r"\bPEP ([0-9]{1,4})\b")            # pep link
re_rfc = re.compile(r"\b(RFC )([0-9]+)\b")              # rfc link


class Fn:
    """Support class for naming in module."""
    def __init__(self, name):
        self.name = name

    def __repr__(self):
        return self.name


def local_name(name):
    """Returns striped name from its parent (module or class).

    Typical use:

        #!jinja
        {{ local_name('MyClass.__init__') }} {# put __init__ to document #}
    """
    dot = name.rfind('.')
    return name[dot+1:]


def property_info(info, delimiter=' | '):
    """Returns property info from tupple.

    Input tuple must containts flags if property is writable, readable
    and deletable.

        #!jinja
        {# return someone like this: WRITE | READ | DELETE #}
        {{ property_info(info) }}
    """
    rv = []
    if info[0]:
        rv.append('READ')
    if info[1]:
        rv.append('WRITE')
    if info[2]:
        rv.append('DELETE')
    return delimiter.join(rv)


def key_doc(a):
    """Return string sortable item via ordering keys."""
    o = ordering[a[0]]
    return str(o[0])+a[1].replace('.', str(o[1]))


def _pep(obj):
    return '<a href="http://www.python.org/dev/peps/pep-{0:04}/">PEP {0}</a>'.\
        format(int(obj.groups()[0]))


def pep_rfc(doc):
    """Automatic create html links in doc from PEP and RFC notifications."""
    doc = re_pep.sub(_pep, doc)
    doc = re_rfc.sub(           # rfc
        r'<a href="http://www.faqs.org/rfcs/rfc\2/">\1\2</a>', doc)
    return doc


class ApiDoc(object):

    def __init__(self):
        self.re_docs = None
        self.api_url = ''
        self.api_keywords = {}
        self.paths = []
        self.modules = []
        self.encoding = 'utf-8'

    def __keyword(self, obj):
        groups = obj.groups()
        key = groups[1]
        args = self.api_keywords[key]
        tmp = '<a href="%s#%s" title="%s">%s</a>' % \
              (self.api_url, key, args, key)
        return groups[0] + tmp + groups[2]

    def __not_in_link(self, obj):
        groups = obj.groups()
        if not groups[0]:
            return groups[1]
        tmp = self.re_docs.sub(self.__keyword, groups[0])
        return tmp + groups[1]

    def uni(self, text):
        """Function always return unicode in Python 2 or str in Python 3."""
        if unicode_exist and isinstance(text, str):
            return text.decode(self.encoding)
        return text

    def linked_api(self, doc):
        """Append link to api to html.

        Function is called by Context.wiki or Context.rst.
        """
        if self.re_docs is not None:
            return re_notlink.sub(self.__not_in_link, doc)
        else:
            return doc

    def load_module(self, module):               # jinja function
        """Get documentation of function, variables and classes from module.

        Example:

            #!jinja
            {% set api = load_module('module') %}
            {% for type, name, args, doc = api[0] %}
                ...
        """
        try:
            ml = module.split('.')
            module = __import__(module)
            for it in ml[1:]:                   # if module is submodule
                module = module.__getattribute__(it)
        except Exception as e:
            sys.stderr.write(repr(e)+'\n')
            return []

        if '__file__' in module.__dict__:
            _file = module.__file__
        else:
            _file = module.__name__

        if _file in self.modules:
            return []
        else:
            self.modules.append(_file)

        try:
            source = getsource(module)
        except:
            source = ''

        dependences = set()
        doc = []
        modinfo = {'author': None, 'date': None, 'version': None}

        for name, item in getmembers(module):
            if isclass(item):                       # module classes
                if module.__name__ != item.__module__:
                    dependences.add(item.__module__)
                    continue
                # type, name, args, doc
                doc.append(('class',
                            self.uni(item.__name__),
                            None,
                            self.uni(getdoc(item) or '')))

                for nm, it in getmembers(item):     # class members
                    if ismethod(it) or isfunction(it) \
                            or ismethoddescriptor(it) \
                            or isgetsetdescriptor(it):
                        try:
                            mtype = 'descriptor'
                            if sys.version_info[0] > 2:     # python 3.x
                                if isfunction(it):
                                    mtype = 'method'
                                elif isinstance(it, staticmethod):
                                    mtype = 'staticmethod'
                            else:                           # python 2.x
                                if ismethod(it):
                                    mtype = 'method'
                                elif isfunction(it):
                                    mtype = 'staticmethod'

                            if isinstance(it, staticmethod):  # python 3.x
                                it = it.__func__

                            args, vargs, kwords, defaults = getargspec(it)
                            if defaults:
                                # transport functions to their names
                                ndefaults = []
                                for d in defaults:
                                    if isfunction(d):
                                        ndefaults.append(
                                            Fn(".".join((d.__module__,
                                                         d.__name__))))
                                    elif ismethod(d):
                                        ndefaults.append(
                                            Fn(".".join((d.__objclass__,
                                                         d.__name__))))
                                    else:
                                        ndefaults.append(d)
                                defaults = ndefaults

                            # type, name, args, doc
                            doc.append((
                                mtype,
                                (self.uni(item.__name__) + '.' +
                                 self.uni(it.__name__)),
                                formatargspec(args, vargs, kwords,
                                              defaults),
                                getdoc(it) or ''))
                        except:
                            pass

                    elif isinstance(it, property):
                        # type, name, property info, doc
                        doc.append((
                            'property',
                            self.uni(item.__name__) + '.' + self.uni(nm),
                            (bool(it.fget), bool(it.fset), bool(it.fdel)),
                            getdoc(it) or ''))
                    # elif isbuiltin(it):    <- __new__, __subclasshook__
                    else:
                        pass

            # module function
            elif isfunction(item) or isroutine(item):
                if module.__name__ != item.__module__:
                    item__module__ = item.__module__
                    if item.__module__ is None:
                        item__module__ = "None"
                    elif item.__module__.startswith(module.__name__):
                        item__module__ = \
                            item__module__[len(module.__name__)+1:]
                    # append to dependences
                    dependences.add(item__module__)
                    continue

                if not isbuiltin(item):
                    args, vargs, kwords, defaults = getargspec(item)
                    if defaults:
                        # transport functions to their names
                        ndefaults = []
                        for d in defaults:
                            if isfunction(d):
                                ndefaults.append(
                                    Fn(".".join((d.__module__,
                                                 d.__name__))))
                            elif ismethod(d):
                                ndefaults.append(
                                    Fn(".".join((d.__objclass__,
                                                 d.__name__))))
                            else:
                                ndefaults.append(d)
                        defaults = ndefaults

                    # type, name, args, doc
                    doc.append(('function',
                                name,
                                formatargspec(args, vargs, kwords,
                                              defaults),
                                getdoc(item) or ''))
                else:
                    bdoc = getdoc(item) or ''
                    if bdoc.find(name) == 0:
                        eol = bdoc.find('\n')
                        args = bdoc[len(name):eol]
                        bdoc = bdoc[eol:].strip()
                    else:
                        args = ''

                    # type, name, args, doc
                    doc.append(('function', name, args, bdoc))

            elif ismodule(item):
                doc.append(('submodule', name, None, ''))
            elif isinstance(item, Pattern):
                if name[0:2] == '__' \
                        or not re.search("\n%s\s*=" % name, source):
                    continue
                match = re.search("\n#\s*([^\n]*)\n%s\s*=" % name, source)
                comment = match.groups()[0] if match else ''
                # type, name, value, doc
                doc.append(('variable',
                            name,
                            "(%s, %s)" % (item.pattern, item.flags),
                            comment))
            else:
                if name[0:2] == '__' \
                        or not re.search("\n%s\s*=" % name, source):
                    if name in ('__author__', '__date__', '__version__'):
                        modinfo[name.strip('_')] = repr(item)
                    continue
                # get last previous comment start with hash char (#)
                match = re.search("\n#\s*([^\n]*)\n%s\s*=" % name, source)
                comment = match.groups()[0] if match else ''
                # type, name, value, doc
                doc.append(('variable', name, repr(item), comment))
        # endfor

        doc.append(('module',
                    module.__name__,
                    (modinfo['author'], modinfo['date'],
                     modinfo['version']),
                    getdoc(module) or ''))

        for name in dependences:
            doc.append(('dependence', name, None, ''))

        return sorted(doc, key=key_doc)
    # enddef loadmodule

    def keywords(self, api, api_url="",                 # jinja function
                 types=('module', 'class', 'method', 'staticmethod',
                        'descriptor', 'property', 'variable', 'function',
                        'h1', 'h2', 'h3')):
        """
        Fill internal api_url variable from names of modules, classes,
        functions, methods, variables or h1, h2 and h3 sections. With this,
        wiki can create links to this functions  or classes.

            #!jinja
            {% set api = load_module('module') %}

            {# create api rexex for from module where api will be on
               module_api.html #}
            {% do keywords(api, 'module_api.html') %}

            {# create api rexex from module where api will be on same page #}
            {% do keywords(api) %}

            {# another way how to call keywords function without do keyword #}
            {{ keywords(api) }}    {# keywords function return empty string #}

        Nice to have: there could be nice, if will be arguments in title, so
        mouseover event show some little detail of functions, methods or
        variables.
        """
        api = sorted(api, key=itemgetter(1), reverse=True)
        # TODO: dict for mapping names to parametres (for title in href)

        self.re_docs = re.compile(
            r"(\b)(%s)(\b)" % '|'.join(
                (name.replace('.', '\.')
                    for type, name, args, doc in api if type in types)))
        if self.re_docs.pattern == r"(\b)()(\b)":  # no keywrods found
            self.re_docs = None
        self.api_url = api_url

        self.api_keywords = dict(
            (name, (args if type != 'module' else None) or type)
            for type, name, args, doc in api)

        return ''
    # enddef


# endclass



#!/usr/bin/python

from distutils.core import Command
from distutils.command.build_scripts import build_scripts
from distutils.command.build import build
from distutils.command.clean import clean
from distutils.dir_util import remove_tree
from distutils import log

from os import path, walk
from shutil import copyfile

from setuptools import setup

from jinja24doc import __version__


def doc():
    """Return README.rst content."""
    with open("README.rst", "r") as readme:
        return readme.read().strip()


def find_data_files(directory, target_folder=""):
    """Generate tuple for setup data_files argument."""
    rv = []
    for root, dirs, files in walk(directory):
        if target_folder:
            rv.append((target_folder+root[len(directory):],
                       list(root + '/' + f
                            for f in files if f[0] != '.' and f[-1] != '~')))
        else:
            rv.append((root,
                       list(root + '/' + f
                            for f in files if f[0] != '.' and f[-1] != '~')))
    return rv


class build_doc(Command):
    description = "build html documentation"
    user_options = [
        ('build-base=', 'b',
         "base build directory (default: 'build.build-base')"),
        ('build-doc=', 'd', "directory to build html documentation to"),
        ('public', 'p', "build as part of public poorhttp web")
    ]

    def initialize_options(self):
        self.build_base = None
        self.build_doc = None
        self.public = False

    def finalize_options(self):
        self.set_undefined_options('build',
                                   ('build_base', 'build_base'))
        if self.build_doc is None:
            self.build_doc = path.join(self.build_base, 'doc')

    def run(self):
        from jinja24doc.context import Context
        import sys
        sys.path.insert(0, 'jinja24doc')
        log.info("creating documentation")

        def _jinja24doc(tname):
            ctx = Context(('templates/jinja24doc', 'doc'))
            log.info('jinja24doc processing %s ...' % tname)
            data = ctx.generate(tname, public=self.public)
            if not isinstance(data, str):
                data = data.encode('utf-8')
            with open(self.build_doc+'/'+tname, 'w') as out:
                out.write(data)

        log.info("building html documentation")
        if self.public:
            log.info("building as public part of poorhttp web")
        if self.dry_run:
            return
        if self.build_doc is None:
            return

        self.mkpath(self.build_doc)
        _jinja24doc('index.html')
        _jinja24doc('tools.html')
        _jinja24doc('templates.html')
        _jinja24doc('api.html')
        _jinja24doc('licence.html')
        copyfile('templates/green.css', self.build_doc+'/style.css')
        copyfile('templates/jinja24doc/web.css', self.build_doc+'/web.css')
        copyfile('templates/jinja24doc//small-logo.png',
                 self.build_doc+'/small-logo.png')


class X_build(build):
    sub_commands = build.sub_commands + [('build_doc', lambda self: True)]


class X_clean(clean):
    user_options = clean.user_options + [
        ('build-doc=', 'd', "directory to build html documentation to"),
    ]

    def initialize_options(self):
        clean.initialize_options(self)
        self.build_doc = None

    def finalize_options(self):
        clean.finalize_options(self)
        if self.build_doc is None:
            self.build_doc = path.join(self.build_base, 'doc')

    def run(self):
        if self.all:
            for directory in (self.build_doc, 'build/_scripts_'):
                if path.exists(directory):
                    remove_tree(directory, dry_run=self.dry_run)
                else:
                    log.warn("'%s' does not exist -- can't clean it",
                             directory)
        clean.run(self)


setup(
    name="jinja24doc",
    version=__version__,
    description="Jinja24Doc for Python",
    author="Ondrej Tuma",
    author_email="mcbig@zeropage.cz",
    url="http://poorhttp.zeropage.cz/jinja24doc",
    packages=['jinja24doc'],
    data_files=(
        [('share/doc/jinja24doc', ['LICENCE', 'README.rst'])] +
        find_data_files('doc', 'share/doc/jinja24doc') +
        find_data_files('build/doc', 'share/doc/jinja24doc/html') +
        find_data_files('templates', 'share/jinja24doc/templates')),
    license="BSD",
    long_description=doc(),
    classifiers=[
        "Development Status :: 5 - Production/Stable",
        "Environment :: Console",
        "Intended Audience :: Developers",
        "License :: OSI Approved :: BSD License",
        "Natural Language :: English",
        "Natural Language :: Czech",
        "Operating System :: POSIX",
        "Programming Language :: Python :: 2",
        "Programming Language :: Python :: 3",
        "Topic :: Documentation",
        "Topic :: Software Development :: Documentation",
        "Topic :: Text Processing :: Markup",
        "Topic :: Utilities"
    ],
    install_requires=['jinja2 >= 2.10', 'docutils-tinyhtmlwriter'],
    test_suite='tests',
    cmdclass={'build': X_build,
              'clean': X_clean,
              'build_doc': build_doc},
    entry_points={
        'console_scripts': [
            'jinja24doc = jinja24doc.tools:jinja24doc',
            'rst24doc = jinja24doc.tools:rst24doc',
            'wiki24doc = jinja24doc.tools:wiki24doc',
        ]
    }
)



"""Command tools entry points."""
from jinja24doc.frontend import jinja_cmdline, auto_cmdline


def jinja24doc():
    """Generates documentation output from jinja templates."""
    return jinja_cmdline(description=jinja24doc.__doc__)


def rst24doc():
    """Generates HTML documentation from standalone reStructuredText,
       python file or python module.
    """
    return auto_cmdline(description=rst24doc.__doc__,
                        file_types=('.txt', '.rst'))


def wiki24doc():
    """Generates HTML documentation from standalone wiki,
       python file or python module.
    """
    return auto_cmdline(description=wiki24doc.__doc__, formater='wiki',
                        file_types=('.txt'))
